//--------------------------------------------------------------------------------
// This file is part of the Windows Workflow Foundation Sample Code
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
// This source code is intended only as a supplement to Microsoft
// Development Tools and/or on-line documentation.  See these other
// materials for detailed information regarding Microsoft code samples.
// 
// THIS CODE AND INFORMATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY
// KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//--------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Workflow.ComponentModel;
using System.Workflow.ComponentModel.Compiler;
using System.Workflow.Activities;

namespace Microsoft.Samples.Workflow.UI
{
    class NavigatorValidator : CompositeActivityValidator
    {
        public override ValidationErrorCollection Validate(ValidationManager manager, object obj)
        {
            ValidationErrorCollection errors = base.Validate(manager, obj);

            Navigator nav = obj as Navigator;
            if (nav == null)
            {
                throw new ArgumentException(SR.GetString(SR.Error_UnexpectedArgumentType, typeof(Navigator).FullName), "obj");
            }

            bool foundStartWith = string.IsNullOrEmpty(nav.StartWith);
            foreach (Activity childActivity in nav.EnabledActivities)
            {
                if (!foundStartWith && childActivity.Name == nav.StartWith)
                {
                    foundStartWith = true;
                }

                if (Helpers.IsOOBEventTypeActivity(childActivity.GetType()))
                {
                    errors.Add(new ValidationError(SR.GetString(SR.Error_NoIEventActivity, childActivity.QualifiedName), ErrorNumbers.Error_NoIEventActivity));
                }

                if (Helpers.IsOOBStateMachineActivity(childActivity.GetType()))
                {
                    errors.Add(new ValidationError(SR.GetString(SR.Error_NoStatemachineActivity, childActivity.QualifiedName), ErrorNumbers.Error_NoStatemachineActivity));
                }

                if (Helpers.IsOOBWebServiceActivity(childActivity.GetType()))
                {
                    errors.Add(new ValidationError(SR.GetString(SR.Error_NoWebServiceActivity, childActivity.QualifiedName), ErrorNumbers.Error_NoWebServiceActivity));
                }

                if (childActivity is CompositeActivity && !(childActivity is InteractionBaseActivity))
                {
                    ValidateCompositeChild(errors, (CompositeActivity)childActivity);
                }
            }

            if (!foundStartWith)
            {
                errors.Add(new ValidationError(SR.GetString(SR.Error_InvalidStartWith, nav.QualifiedName), ErrorNumbers.Error_InvalidStartWith, false, "StartWith"));
            }

            return errors;
        }

        static void ValidateCompositeChild(ValidationErrorCollection errors, CompositeActivity compositeActivity)
        {
            int countInteractionActivity = 0;
            Queue<CompositeActivity> compositeActivities = new Queue<CompositeActivity>();
            compositeActivities.Enqueue(compositeActivity);
            while (compositeActivities.Count > 0)
            {
                compositeActivity = compositeActivities.Dequeue();
                foreach (Activity child in compositeActivity.EnabledActivities)
                {
                    if (Helpers.IsOOBEventTypeActivity(child.GetType()))
                    {
                        errors.Add(new ValidationError(SR.GetString(SR.Error_NoIEventActivity, child.QualifiedName), ErrorNumbers.Error_NoIEventActivity));
                    }

                    if (Helpers.IsOOBStateMachineActivity(child.GetType()))
                    {
                        errors.Add(new ValidationError(SR.GetString(SR.Error_NoStatemachineActivity, child.QualifiedName), ErrorNumbers.Error_NoStatemachineActivity));
                    }

                    if (Helpers.IsOOBWebServiceActivity(child.GetType()))
                    {
                        errors.Add(new ValidationError(SR.GetString(SR.Error_NoWebServiceActivity, child.QualifiedName), ErrorNumbers.Error_NoWebServiceActivity));
                    }

                    if (child is InteractionBaseActivity)
                    {
                        countInteractionActivity++;
                    }

                    if (child is CompositeActivity && !(child is InteractionBaseActivity))
                    {
                        compositeActivities.Enqueue(child as CompositeActivity);
                    }
                }
            }

            if (countInteractionActivity > 1)
            {
                errors.Add(new ValidationError(SR.GetString(SR.Error_MultipleInteractiveActivity, compositeActivity.QualifiedName), ErrorNumbers.Error_MultipleInteractiveActivity));
            }

            return;
        }
    }
}
